# Copyright (c) 2014 by Ladislav Lhotka, CZ.NIC <lhotka@nic.cz>
#
# Pyang plugin generating a XSLT1 stylesheet for XML->JSON translation.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""JSONXSL output plugin

This plugin takes a YANG data model and produces an XSLT1 stylesheet
that is able to translate any valid XML instance document or NETCONF
message into JSON.
"""

import os
import sys
import xml.etree.ElementTree as ET

from pyang import plugin, statements, error
from pyang.util import unique_prefixes

ss = ET.Element("stylesheet",
                {"version": "1.0",
                 "xmlns": "http://www.w3.org/1999/XSL/Transform",
                 "xmlns:nc": "urn:ietf:params:xml:ns:netconf:base:1.0",
                 "xmlns:en": "urn:ietf:params:xml:ns:netconf:notification:1.0"})
"""Root element of the output XSLT stylesheet."""

type_class = dict((t,"unquoted") for t in
                  ("boolean", "int8", "int16", "int32",
                   "uint8", "uint16", "uint32"))
"""Classification of types suited for JSON translation."""

type_class.update((t,t) for t in
                  ("empty", "instance-identifier", "identityref", "string"))

union_class = dict((t,"integer") for t in
                   ("int8", "int16", "int32",
                   "uint8", "uint16", "uint32"))
"""Classification of types needed for resolving union-typed values."""

union_class.update({"boolean": "boolean"})

def pyang_plugin_init():
    plugin.register_plugin(JsonXslPlugin())

class JsonXslPlugin(plugin.PyangPlugin):
    def add_output_format(self, fmts):
        self.multiple_modules = True
        fmts['jsonxsl'] = self

    def setup_fmt(self, ctx):
        ctx.implicit_errors = False

    def emit(self, ctx, modules, fd):
        """Main control function.

        Set up the top-level parts of the stylesheet, then process
        recursively all nodes in all data trees, and finally emit the
        serialized stylesheet.
        """
        for (epos, etag, eargs) in ctx.errors:
            if error.is_error(error.err_level(etag)):
                raise error.EmitError("JSONXSL plugin needs a valid module")
        self.real_prefix = unique_prefixes(ctx)
        self.top_names = []
        for m in modules:
            self.top_names.extend([c.arg for c in m.i_children if
                                   c.keyword not in ("rpc", "notification")])
        tree = ET.ElementTree(ss)
        ET.SubElement(ss, "output", method="text")
        xsltdir = os.environ.get("PYANG_XSLT_DIR",
                                 "/usr/local/share/yang/xslt")
        ET.SubElement(ss, "include", href=xsltdir + "/jsonxsl-templates.xsl")
        ET.SubElement(ss, "strip-space", elements="*")
        nsmap = ET.SubElement(ss, "template", name="nsuri-to-module")
        ET.SubElement(nsmap, "param", name="uri")
        choo = ET.SubElement(nsmap, "choose")
        for module in self.real_prefix.keys():
            ns_uri = module.search_one("namespace").arg
            ss.attrib["xmlns:" + self.real_prefix[module]] = ns_uri
            when = ET.SubElement(choo, "when", test="$uri='" + ns_uri + "'")
            self.xsl_text(module.i_modulename, when)
            self.process_module(module)
        if sys.version > "3":
            tree.write(fd, encoding="unicode", xml_declaration=True)
        elif sys.version > "2.7":
            tree.write(fd, encoding="UTF-8", xml_declaration=True)
        else:
            tree.write(fd, encoding="UTF-8")

    def process_module(self, yam):
        """Process data nodes, RPCs and notifications in a single module."""
        for ann in yam.search(("ietf-yang-metadata", "annotation")):
            self.process_annotation(ann)
        for ch in yam.i_children[:]:
            if ch.keyword == "rpc":
                self.process_rpc(ch)
            elif ch.keyword == "notification":
                self.process_notification(ch)
            else:
                continue
            yam.i_children.remove(ch)
        self.process_children(yam, "//nc:*", 1)

    def process_annotation(self, ann):
        """Process metadata annotation."""
        tmpl = self.xsl_template("@" + self.qname(ann))
        ET.SubElement(tmpl, "param", name="level", select="0")
        ct = self.xsl_calltemplate("leaf", tmpl)
        ET.SubElement(ct, "with-param", name="level", select="$level")
        self.xsl_withparam("nsid", ann.i_module.i_modulename + ":", ct)
        self.type_param(ann, ct)

    def process_rpc(self, rpc):
        """Process input and output parts of `rpc`."""
        p = "/nc:rpc/" + self.qname(rpc)
        tmpl = self.xsl_template(p)
        inp = rpc.search_one("input")
        if inp is not None:
            ct = self.xsl_calltemplate("rpc-input", tmpl)
            self.xsl_withparam("nsid", rpc.i_module.i_modulename + ":", ct)
            self.process_children(inp, p, 2)
        outp = rpc.search_one("output")
        if outp is not None:
            self.process_children(outp, "/nc:rpc-reply", 1)

    def process_notification(self, ntf):
        """Process event notification `ntf`."""
        p = "/en:notification/" + self.qname(ntf)
        tmpl = self.xsl_template(p)
        ct = self.xsl_calltemplate("container", tmpl)
        self.xsl_withparam("level", "1", ct)
        if ntf.arg == "eventTime":            # local name collision
            self.xsl_withparam("nsid", ntf.i_module.i_modulename + ":", ct)
        self.process_children(ntf, p, 2)

    def process_children(self, node, path, level, parent=None):
        """Process all children of `node`.

        `path` is the Xpath of `node` which is used in the 'select'
        attribute of XSLT templates.
        """
        data_parent = parent if parent else node
        chs = node.i_children
        for ch in chs:
            if ch.keyword in ["choice", "case"]:
                self.process_children(ch, path, level, node)
                continue
            p = path + "/" + self.qname(ch)
            tmpl = self.xsl_template(p)
            ct = self.xsl_calltemplate(ch.keyword, tmpl)
            self.xsl_withparam("level", "%d" % level, ct)
            if (data_parent.i_module is None or
                ch.i_module.i_modulename != data_parent.i_module.i_modulename):
                self.xsl_withparam("nsid", ch.i_module.i_modulename + ":", ct)
            if ch.keyword in ["leaf", "leaf-list"]:
                self.type_param(ch, ct)
            elif ch.keyword != "anyxml":
                offset = 2 if ch.keyword == "list" else 1
                self.process_children(ch, p, level + offset)

    def type_param(self, node, ct):
        """Resolve the type of a leaf or leaf-list node for JSON.
        """
        types = self.get_types(node)
        ftyp = types[0]
        if len(types) == 1:
            if ftyp in type_class:
                jtyp = type_class[ftyp]
            else:
                jtyp = "other"
            self.xsl_withparam("type", jtyp, ct)
        elif ftyp in ["string", "enumeration", "bits", "binary",
                      "identityref", "instance-identifier"]:
            self.xsl_withparam("type", "string", ct)
        else:
            opts = []
            for t in types:
                if t in union_class:
                    ut = union_class[t]
                elif t in ["int64", "uint64"] or t.startswith("decimal@"):
                    ut = t
                else:
                    ut = "other"
                if ut not in opts:
                    opts.append(ut)
                    if ut == "other": break
                    if ut == "decimal" and "integer" not in opts:
                        opts.append("integer")
            self.xsl_withparam("type", "union", ct)
            self.xsl_withparam("options", ",".join(opts) + ",", ct)

    def get_types(self, node):
        res = []
        def resolve(typ):
            if typ is None:
                res.append("string") # default for annotations
            elif typ.arg == "leafref":
                resolve(typ.i_type_spec.i_target_node.search_one("type"))
            elif typ.arg == "union":
                for ut in typ.i_type_spec.types: resolve(ut)
            elif typ.arg == "decimal64":
                res.append("decimal@" +
                           typ.search_one("fraction-digits").arg)
            elif typ.i_typedef is not None:
                resolve(typ.i_typedef.search_one("type"))
            else:
                res.append(typ.arg)
        resolve(node.search_one("type"))
        return res

    def qname(self, node):
        """Return the qualified name of `node`.

        In JSON, namespace identifiers are YANG module names.
        """
        return self.real_prefix[node.main_module()] + ":" + node.arg

    def xsl_template(self, name):
        """Construct an XSLT 'template' element matching `name`."""
        return ET.SubElement(ss, "template" , match = name)

    def xsl_text(self, text, parent):
        """Construct an XSLT 'text' element containing `text`.

        `parent` is this element's parent.
        """
        res = ET.SubElement(parent, "text")
        res.text = text
        return res

    def xsl_calltemplate(self, name, parent):
        """Construct an XSLT 'call-template' element.

        `parent` is this element's parent.
        `name` is the name of the template to be called.
        """
        return ET.SubElement(parent, "call-template", name=name)

    def xsl_withparam(self, name, value, parent):
        """Construct an XSLT 'with-param' element.

        `parent` is this element's parent.
        `name` is the parameter name.
        `value` is the parameter value.
        """
        res = ET.SubElement(parent, "with-param", name=name)
        res.text = value
        return res



"""Capability URL plugin

"""

import optparse
import sys
import os.path

from pyang import plugin
from pyang import util

def pyang_plugin_init():
    plugin.register_plugin(CapabilityPlugin())

class CapabilityPlugin(plugin.PyangPlugin):
    def add_output_format(self, fmts):
        self.multiple_modules = True
        fmts['capability'] = self
    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--capability-entity",
                                 dest="capa_entity",
                                 action="store_true",
                                 default=False,
                                 help="Write ampersands as XML entity")
            ]
        g = optparser.add_option_group("Capability output specific options")
        g.add_options(optlist)
    def emit(self, ctx, modules, fd):
        for m in modules:
            emit_capability(ctx, m, fd)

def emit_capability(ctx, m, fd):
    amp = "&amp;" if ctx.opts.capa_entity else "&"
    ns = m.search_one('namespace')
    if ns is None:
        return
    s = ns.arg + "?module=" + m.i_modulename

    latest_rev = util.get_latest_revision(m)
    if latest_rev != "unknown":
        s = s + amp + "revision=" + latest_rev

    if m.i_modulename in ctx.features:
        if len(ctx.features[m.i_modulename]) > 0:
            s = s + amp + "features=" + ",".join(ctx.features[m.i_modulename])
        else:
            # do not report any features from the module
            pass
    else:
        # report all features defined in the module
        fs = [x.arg for x in m.search('feature')]
        s = s + amp + "features=" + ",".join(fs)

    devs = []
    for d in ctx.deviation_modules:
        # check if this deviation module deviates anything in our module
        for dev in d.search('deviation'):
            if (dev.i_target_node is not None and
                dev.i_target_node.i_module.i_modulename == m.i_modulename):
                devs.append(d.i_modulename)
                break

    if len(devs) > 0:
        s = s + amp + "deviations=" +  ",".join(devs)

    fd.write(s + '\n')



"""pyang plugin handling"""

import os
import sys

plugins = []
"""List of registered PyangPlugin instances"""

def init(plugindirs=[]):
    """Initialize the plugin framework"""

    # initialize the builtin plugins
    from .translators import yang,yin,dsdl
    yang.pyang_plugin_init()
    yin.pyang_plugin_init()
    dsdl.pyang_plugin_init()

    # search for plugins in std directory
    basedir = os.path.split(sys.modules['pyang'].__file__)[0]
    plugindirs.insert(0, basedir + "/plugins")

    # add paths from env
    pluginpath = os.getenv('PYANG_PLUGINPATH')
    if pluginpath is not None:
        plugindirs.extend(pluginpath.split(os.pathsep))

    syspath = sys.path
    for plugindir in plugindirs:
        sys.path = [plugindir] + syspath
        try:
            fnames = os.listdir(plugindir)
        except OSError:
            continue
        for fname in fnames:
            if not fname.startswith(".#") and fname.endswith(".py") and \
               fname != '__init__.py':
                pluginmod = __import__(fname[:-3])
                try:
                    pluginmod.pyang_plugin_init()
                except AttributeError as s:
                    print(pluginmod.__dict__)
                    raise AttributeError(pluginmod.__file__ + ': ' + str(s))
        sys.path = syspath

def register_plugin(plugin):
    """Call this to register a pyang plugin. See class PyangPlugin
    for more info.
    """
    plugins.append(plugin)

class PyangPlugin(object):
    """Abstract base class for pyang plugins

    A pyang plugin is a module found in the plugins directory of the
    pyang installation, or in the dynamic pluginpath.

    Such a module must export a function 'pyang_plugin_init()', which
    may call pyang.plugin.register_plugin() with an instance of a class
    derived from this class as argument.

    A plugin can extend the base pyang library functions, or the pyang
    front-end program, or both.
    """

    def __init__(self):
        self.multiple_modules = False
        self.handle_comments = False

    ## pyang front-end program methods

    def add_output_format(self, fmts):
        """Add an output format to the pyang program.

        `fmts` is a dict which maps the format name string to a plugin
        instance.

        Override this method and update `fmts` with the output format
        name.
        """
        return
    def add_opts(self, optparser):
        """Add command line options to the pyang program.

        Override this method and add the plugin related options as an
        option group.
        """
        return

    ## library methods

    def setup_ctx(self, ctx):
        """Modify the Context at setup time.  Called for all plugins.

        Override this method to modify the Context before the module
        repository is accessed.
        """
        return

    def setup_fmt(self, ctx):
        """Modify the Context at setup time.  Called for the selected plugin.

        Override this method to modify the Context before the module
        repository is accessed.
        """
        return

    def pre_load_modules(self, ctx):
        """Called for the selected plugin, before any modules are loaded"""
        return

    def pre_validate_ctx(self, ctx, modules):
        """Called for all plugins, before the modules are validated"""
        return

    def pre_validate(self, ctx, modules):
        """Called for the selected plugin, before the modules are validated"""
        return

    def post_validate(self, ctx, modules):
        """Called for the selected plugin, after the modules
        have been validated"""
        return

    def post_validate_ctx(self, ctx, modules):
        """Called for all plugins, after the modules
        have been validated"""
        return

    def emit(self, ctx, modules, writef):
        """Produce the plugin output.

        Override this method to perform the output conversion.
        `writef` is a function that takes one string to print as argument.

        Raise error.EmitError on failure.
        """
        return



import re
import sys

# not 100% XPath / XML, but good enough for YANG
namestr=r'[a-zA-Z_][a-zA-Z0-9_\-.]*'
ncnamestr = '((' + namestr + '):)?(' + namestr + ')'
prefixteststr = '((' + namestr + r'):)?\*'

patterns = [
    ('whitespace', re.compile(r'\s+')),
    # Expr tokens
    ('(', re.compile(r'\(')),
    (')', re.compile(r'\)')),
    ('[', re.compile(r'\[')),
    (']', re.compile(r'\]')),
    ('..', re.compile(r'\.\.')),
    ('.', re.compile(r'\.')),
    ('@', re.compile(r'\@')),
    (',', re.compile(r',')),
    ('::', re.compile(r'::')),
    # operators
    ('//', re.compile(r'\/\/')),
    ('/', re.compile(r'\/')),
    ('|', re.compile(r'\|')),
    ('+', re.compile(r'\+')),
    ('-', re.compile(r'-')),
    ('=', re.compile(r'=')),
    ('!=', re.compile(r'!=')),
    ('<=', re.compile(r'<=')),
    ('>=', re.compile(r'>=')),
    ('>', re.compile(r'>')),
    ('<', re.compile(r'<')),
    ('*', re.compile(r'\*')),
    # others
    ('number', re.compile(r'[0-9]+(\.[0-9]+)?')),
    ('prefix-test', re.compile(prefixteststr)),
    ('name', re.compile(ncnamestr)),
    ('attribute', re.compile(r'\@' + ncnamestr)),
    ('variable', re.compile(r'\$' + ncnamestr)),
    ('literal', re.compile(r'(\".*?\")|(\'.*?\')')),
    ]

operators = [ 'div', 'and', 'or', 'mod' ]
node_types = [ 'comment', 'text', 'processing-instruction', 'node' ]
axes = [ 'ancestor-or-self', 'ancestor', 'attribute', 'child',
         'descendant-or-self', 'descendant', 'following-sibling',
         'following', 'namespace', 'parent', 'preceding-sibling',
         'preceding', 'self' ]

re_open_para = re.compile(r'\s*\(')
re_axis = re.compile(r'\s*::')

def validate(s):
    """Validate the XPath expression in the string `s`
    Return True if the expression is correct, and throw
    SyntaxError on failure."""
    t = tokens(s)
    return True

def tokens(s):
    """Return a list of tokens, or throw SyntaxError on failure.
    A token is one of the patterns or:
      ('wildcard', '*')
      ('axis', axisname)
    """
    pos = 0
    toks = []
    while pos < len(s):
        matched = False
        for (tokname, r) in patterns:
            m = r.match(s, pos)
            if m is not None:
                # found a matching token
                prec = _preceding_token(toks)
                if tokname == '*' and prec is not None and _is_special(prec):
                    # XPath 1.0 spec, 3.7 special rule 1a
                    # interpret '*' as a wildcard
                    tok = ('wildcard', m.group(0))
                elif (tokname == 'name' and
                      prec is not None and not _is_special(prec) and
                      m.group(0) in operators):
                    # XPath 1.0 spec, 3.7 special rule 1b
                    # interpret the name as an operator
                    tok = (m.group(0), m.group(0))
                elif tokname == 'name':
                    # check if next token is '('
                    if re_open_para.match(s, pos + len(m.group(0))):
                        # XPath 1.0 spec, 3.7 special rule 2
                        if m.group(0) in node_types:
                            # XPath 1.0 spec, 3.7 special rule 2a
                            tok = (m.group(0), m.group(0))
                        else:
                            # XPath 1.0 spec, 3.7 special rule 2b
                            tok = ('function', m.group(0))
                    # check if next token is '::'
                    elif re_axis.match(s, pos + len(m.group(0))):
                        # XPath 1.0 spec, 3.7 special rule 3
                        if m.group(0) in axes:
                            tok = ('axis', m.group(0))
                        else:
                            e = "%s: unknown axis %s" % (pos+1, m.group(0))
                            raise SyntaxError(e)
                    else:
                        tok = ('name', m.group(0))
                else:
                    tok = (tokname, m.group(0))
                pos += len(m.group(0))
                toks.append(tok)
                matched = True
                break
        if matched == False:
            # no patterns matched
            raise SyntaxError("at position %s" % str(pos+1))
    return toks

def _preceding_token(toks):
    if len(toks) > 1 and toks[-1][0] == 'whitespace':
        return toks[-2][0]
    if len(toks) > 0 and toks[-1][0] != 'whitespace':
        return toks[-1][0]
    return None

_special_toks = [ ',', '@', '::', '(', '[', '/', '//', '|', '+', '-',
                 '=', '!=', '<', '<=', '>', '>=',
                 'and', 'or', 'mod', 'div' ]

def _is_special(tok):
    return tok in _special_toks


def add_prefix(prefix, s):
    "Add `prefix` to all unprefixed names in `s`"
    # tokenize the XPath expression
    toks = tokens(s)
    # add default prefix to unprefixed names
    toks2 = [_add_prefix(prefix, tok) for tok in toks]
    # build a string of the patched expression
    ls = [x for (_tokname, x) in toks2]
    return ''.join(ls)

_re_ncname = re.compile(ncnamestr)
def _add_prefix(prefix, tok):
    (tokname, s) = tok
    if tokname == 'name':
        m = _re_ncname.match(s)
        if m.group(2) == None:
            return (tokname, prefix + ':' + s)
    return tok

core_functions = [
    'last',
    'position',
    'count',
    'id',
    'local-name',
    'namespace-uri',
    'name',
    'string',
    'concat',
    'starts-with',
    'contains',
    'substring-before',
    'substring-after',
    'substring',
    'string-length',
    'normalize-space',
    'translate',
    'boolean',
    'not',
    'true',
    'false',
    'lang',
    'number',
    'sum',
    'floor',
    'ceiling',
    'round',
    ]



"""Package pyang.translators: output or translators plugins for YANG.

Modules:

* dsdl: DSDL translator plugin
* rng: RELAX NG translator plugin 
* xsd: W3C XML Schema translator plugin
* yin: YIN output plugin
* yang: YANG output plugin
* tree: Tree output plugin

"""



from setuptools import setup
from setuptools import Distribution
from os.path import join
import pyang
import glob
import os
import re
import sys

modules_iana = glob.glob(os.path.join('modules', 'iana', '*.yang'))
modules_ietf = glob.glob(os.path.join('modules', 'ietf', '*.yang'))
xslt = glob.glob(os.path.join('xslt', '*.xsl'))
schema = glob.glob(os.path.join('schema', '*.rng'))
images = glob.glob(os.path.join('tools', 'images', '*'))
man1 = glob.glob(os.path.join('man', 'man1', '*.1'))

class PyangDist(Distribution):

      """The purpose of this subclass of Distribution is to extend the
      install procedure with preprocessing of shell scripts and man
      pages so that they reflect the actual installation prefix, which
      may be changed through the --prefix option.
      """

      def preprocess_files(self, prefix):
            """Change the installation prefix where necessary.
            """
            if prefix is None: return
            files = ("bin/yang2dsdl", "man/man1/yang2dsdl.1",
                     "pyang/plugins/jsonxsl.py")
            regex = re.compile("^(.*)/usr/local(.*)$")
            for f in files:
                  inf = open(f)
                  cnt = inf.readlines()
                  inf.close()
                  ouf = open(f,"w")
                  for line in cnt:
                        mo = regex.search(line)
                        if mo is None:
                              ouf.write(line)
                        else:
                              ouf.write(mo.group(1) + prefix + mo.group(2) +
                                        "\n")
                  ouf.close()

      def run_commands(self):
            opts = self.command_options
            if "install" in opts:
                  self.preprocess_files(opts["install"].get("prefix",
                                                            ("", None))[1])
            Distribution.run_commands(self)

# If the installation is on windows, place pyang.bat file in Scripts directory
if os.sep == '\\':
    script_dir = join(sys.prefix, 'Scripts')
    pyang_file = join(script_dir, 'pyang')
    path = join(script_dir, 'pyang.bat')
    with open(path, 'w') as script:
        script.write('@echo off\n%s %s %%*\n' % ('python', pyang_file))

setup(name='pyang',
      version=pyang.__version__,
      author='Martin Bjorklund',
      author_email='mbj@tail-f.com',
      description="A YANG (RFC 6020) validator and converter",
      long_description="An extensible  YANG (RFC 6020) validator.  Provides a framwork for plugins that can convert YANG modules to other formats.",
      url='https://github.com/mbj4668/pyang',
      license='BSD',
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            ],
      keywords='YANG validator',
      distclass=PyangDist,
      scripts=['bin/pyang', 'bin/yang2html', 'bin/yang2dsdl', 'bin/json2xml'],
      packages=['pyang', 'pyang.plugins', 'pyang.translators'],
      data_files=[
            ('share/man/man1', man1),
            ('share/yang/modules/iana', modules_iana),
            ('share/yang/modules/ietf', modules_ietf),
            ('share/yang/xslt', xslt),
            ('share/yang/images', images),
            ('share/yang/schema', schema),
            ('etc/bash_completion.d', ['etc/bash_completion.d/pyang']),
            ]
      )



"""YIN output plugin"""

from xml.sax.saxutils import quoteattr
from xml.sax.saxutils import escape

import optparse
import re

from .. import plugin
from .. import util
from .. import grammar
from .. import syntax
from .. import statements

yin_namespace = "urn:ietf:params:xml:ns:yang:yin:1"

def pyang_plugin_init():
    plugin.register_plugin(YINPlugin())

class YINPlugin(plugin.PyangPlugin):
    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--yin-canonical",
                                 dest="yin_canonical",
                                 action="store_true",
                                 help="Print in canonical order"),
            optparse.make_option("--yin-pretty-strings",
                                 dest="yin_pretty_strings",
                                 action="store_true",
                                 help="Pretty print strings"),
            ]
        g = optparser.add_option_group("YIN output specific options")
        g.add_options(optlist)
    def add_output_format(self, fmts):
        fmts['yin'] = self
    def emit(self, ctx, modules, fd):
        module = modules[0]
        emit_yin(ctx, module, fd)
        
def emit_yin(ctx, module, fd):
    fd.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    fd.write('<%s name="%s"\n' % (module.keyword, module.arg))
    fd.write(' ' * len(module.keyword) + '  xmlns="%s"' % yin_namespace)

    prefix = module.search_one('prefix')
    if prefix is not None:
        namespace = module.search_one('namespace')
        fd.write('\n')
        fd.write(' ' * len(module.keyword))
        fd.write('  xmlns:' + prefix.arg + '=' +
                 quoteattr(namespace.arg))
    else:
        belongs_to = module.search_one('belongs-to')
        if belongs_to is not None:
            prefix = belongs_to.search_one('prefix')
            if prefix is not None:
                # read the parent module in order to find the namespace uri
                res = ctx.read_module(belongs_to.arg, extra={'no_include':True})
                if res is not None:
                    namespace = res.search_one('namespace')
                    if namespace is None or namespace.arg is None:
                        pass
                    else:
                        # success - namespace found
                        fd.write('\n')
                        fd.write(' ' * len(module.keyword))
                        fd.write('  xmlns:' + prefix.arg + '=' +
                                 quoteattr(namespace.arg))
            
    for imp in module.search('import'):
        prefix = imp.search_one('prefix')
        if prefix is not None:
            rev = None
            r = imp.search_one('revision-date')
            if r is not None:
                rev = r.arg
            mod = statements.modulename_to_module(module, imp.arg, rev)
            if mod is not None:
                ns = mod.search_one('namespace')
                if ns is not None:
                    fd.write('\n')
                    fd.write(' ' * len(module.keyword))
                    fd.write('  xmlns:' + prefix.arg + '=' +
                             quoteattr(ns.arg))
    fd.write('>\n')
    if ctx.opts.yin_canonical:
        substmts = grammar.sort_canonical(module.keyword, module.substmts)
    else:
        substmts = module.substmts
    for s in substmts:
        emit_stmt(ctx, module, s, fd, '  ', '  ')
    fd.write('</%s>\n' % module.keyword)
    
def emit_stmt(ctx, module, stmt, fd, indent, indentstep):
    if util.is_prefixed(stmt.raw_keyword):
        # this is an extension.  need to find its definition
        (prefix, identifier) = stmt.raw_keyword
        tag = prefix + ':' + identifier
        if stmt.i_extension is not None:
            ext_arg = stmt.i_extension.search_one('argument')
            if ext_arg is not None:
                yin_element = ext_arg.search_one('yin-element')
                if yin_element is not None and yin_element.arg == 'true':
                    argname = prefix + ':' + ext_arg.arg
                    argiselem = True
                else:
                    # explicit false or no yin-element given
                    argname = ext_arg.arg
                    argiselem = False
            else:
                argiselem = False
                argname = None
        else:
            argiselem = False
            argname = None
    else:
        (argname, argiselem) = syntax.yin_map[stmt.raw_keyword]
        tag = stmt.raw_keyword
    if argiselem == False or argname is None:
        if argname is None:
            attr = ''
        else:
            attr = ' ' + argname + '=' + quoteattr(stmt.arg)
        if len(stmt.substmts) == 0:
            fd.write(indent + '<' + tag + attr + '/>\n')
        else:
            fd.write(indent + '<' + tag + attr + '>\n')
            for s in stmt.substmts:
                emit_stmt(ctx, module, s, fd, indent + indentstep,
                          indentstep)
            fd.write(indent + '</' + tag + '>\n')
    else:
        fd.write(indent + '<' + tag + '>\n')
        if ctx.opts.yin_pretty_strings:
            # since whitespace is significant in XML, the current
            # code is strictly speaking incorrect.  But w/o the whitespace,
            # it looks too ugly.
            fd.write(indent + indentstep + '<' + argname + '>\n')
            fd.write(fmt_text(indent + indentstep + indentstep, stmt.arg))
            fd.write('\n' + indent + indentstep + '</' + argname + '>\n')
        else:
            fd.write(indent + indentstep + '<' + argname + '>' + \
                       escape(stmt.arg) + \
                       '</' + argname + '>\n')
        if ctx.opts.yin_canonical:
            substmts = grammar.sort_canonical(stmt.keyword, stmt.substmts)
        else:
            substmts = stmt.substmts
        for s in substmts:
            emit_stmt(ctx, module, s, fd, indent + indentstep, indentstep)
        fd.write(indent + '</' + tag + '>\n')

def fmt_text(indent, data):
    res = []
    for line in re.split("(\n)", escape(data)):
        if line == '':
            continue
        if line == '\n':
            res.extend(line)
        else:
            res.extend(indent + line)
    return ''.join(res)



"""YANG usage guidelines plugin
See RFC 6087
Other plugins can derive from this and make it more specific, e.g.,
ietf.py derives from this and sets the namespace and module name prefixes
to IETF-specific values.
"""

import optparse
import sys

from pyang import plugin
from pyang import statements
from pyang import error
from pyang.error import err_add

def pyang_plugin_init():
    plugin.register_plugin(LintPlugin())

class LintPlugin(plugin.PyangPlugin):
    def __init__(self):
        plugin.PyangPlugin.__init__(self)
        ## Other plugins (e.g., ietf.py) can derive from this plugin
        ## and override these variables.

        # Set this to a list of allowed namespace prefixes.
        # The code checks that the namespace is on the form:
        #   <prefix><modulename>
        # If some other convention is used, the derived plugin can
        # define its own checks.
        self.namespace_prefixes = []

        # Set this to a list of allowed module name prefixes.
        # The code checks that the module name is on the form:
        #   <prefix>-...
        # If some other convention is used, the derived plugin can
        # define its own checks.
        self.modulename_prefixes = []

    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--lint",
                                 dest="lint",
                                 action="store_true",
                                 help="Validate the module(s) according to " \
                                 "RFC 6087 rules."),
            optparse.make_option("--lint-namespace-prefix",
                                 dest="lint_namespace_prefixes",
                                 default=[],
                                 action="append",
                                 help="Validate that the module's namespace " \
                                     "matches one of the given prefixes."),
            optparse.make_option("--lint-modulename-prefix",
                                 dest="lint_modulename_prefixes",
                                 default=[],
                                 action="append",
                                 help="Validate that the module's name " \
                                     "matches one of the given prefixes."),
            ]
        optparser.add_options(optlist)

    def setup_ctx(self, ctx):
        if not ctx.opts.lint:
            return
        self._setup_ctx(ctx)

    def _setup_ctx(self, ctx):
        "Should be called by any derived plugin's setup_ctx() function."

        ctx.strict = True
        ctx.canonical = True
        ctx.max_identifier_len = 64
        ctx.implicit_errors = False

        # always add additional prefixes given on the command line
        self.namespace_prefixes.extend(ctx.opts.lint_namespace_prefixes)
        self.modulename_prefixes.extend(ctx.opts.lint_modulename_prefixes)

        # register our grammar validation funs

        statements.add_validation_var(
            '$chk_default',
            lambda keyword: keyword in _keyword_with_default)
        statements.add_validation_var(
            '$chk_required',
            lambda keyword: keyword in _required_substatements)

        statements.add_validation_var(
            '$chk_recommended',
            lambda keyword: keyword in _recommended_substatements)

        statements.add_validation_fun(
            'grammar', ['$chk_default'],
            lambda ctx, s: v_chk_default(ctx, s))
        statements.add_validation_fun(
            'grammar', ['$chk_required'],
            lambda ctx, s: v_chk_required_substmt(ctx, s))
        statements.add_validation_fun(
            'grammar', ['$chk_recommended'],
            lambda ctx, s: v_chk_recommended_substmt(ctx, s))

        statements.add_validation_fun(
            'grammar', ['namespace'],
            lambda ctx, s: v_chk_namespace(ctx, s, self.namespace_prefixes))

        statements.add_validation_fun(
            'grammar', ['module', 'submodule'],
            lambda ctx, s: v_chk_module_name(ctx, s, self.modulename_prefixes))

        statements.add_validation_fun(
            'strict', ['include'],
            lambda ctx, s: v_chk_include(ctx, s))

        statements.add_validation_fun(
            'strict', ['module'],
            lambda ctx, s: v_chk_mandatory_top_level(ctx, s))

        # register our error codes
        error.add_error_code(
            'LINT_EXPLICIT_DEFAULT', 4,
            'RFC 6087: 4.3: '
            + 'statement "%s" is given with its default value "%s"')
        error.add_error_code(
            'LINT_MISSING_REQUIRED_SUBSTMT', 3,
            '%s: '
            + 'statement "%s" must have a "%s" substatement')
        error.add_error_code(
            'LINT_MISSING_RECOMMENDED_SUBSTMT', 4,
            '%s: '
            + 'statement "%s" should have a "%s" substatement')
        error.add_error_code(
            'LINT_BAD_NAMESPACE_VALUE', 4,
            'RFC 6087: 4.8: namespace value should be "%s"')
        error.add_error_code(
            'LINT_BAD_MODULENAME_PREFIX_1', 4,
            'RFC 6087: 4.1: '
            + 'the module name should start with the string %s')
        error.add_error_code(
            'LINT_BAD_MODULENAME_PREFIX_N', 4,
            'RFC 6087: 4.1: '
            + 'the module name should start with one of the strings %s')
        error.add_error_code(
            'LINT_NO_MODULENAME_PREFIX', 4,
            'RFC 6087: 4.1: '
            + 'no module name prefix string used')
        error.add_error_code(
            'LINT_BAD_REVISION', 3,
            'RFC 6087: 4.6: '
            + 'the module\'s revision %s is older than '
            + 'submodule %s\'s revision %s')
        error.add_error_code(
            'LINT_TOP_MANDATORY', 3,
            'RFC 6087: 4.9: '
            + 'top-level node %s must not be mandatory')

        # override std error string
        error.add_error_code(
            'LONG_IDENTIFIER', 3,
            'RFC 6087: 4.2: identifier %s exceeds %s characters')

_keyword_with_default = {
    'status': 'current',
    'mandatory': 'false',
    'min-elements': '0',
    'max-elements': 'unbounded',
    'config': 'true',
    'yin-element': 'false',
    }

_required_substatements = {
    'module': (('contact', 'organization', 'description', 'revision'),
               "RFC 6087: 4.7"),
    'submodule': (('contact', 'organization', 'description', 'revision'),
                  "RFC 6087: 4.7"),
    'revision':(('reference',), "RFC 6087: 4.7"),
    'extension':(('description',), "RFC 6087: 4.12"),
    'feature':(('description',), "RFC 6087: 4.12"),
    'identity':(('description',), "RFC 6087: 4.12"),
    'typedef':(('description',), "RFC 6087: 4.11,4.12"),
    'grouping':(('description',), "RFC 6087: 4.12"),
    'augment':(('description',), "RFC 6087: 4.12"),
    'rpc':(('description',), "RFC 6087: 4.12"),
    'notification':(('description',), "RFC 6087: 4.12,4.14"),
    'container':(('description',), "RFC 6087: 4.12"),
    'leaf':(('description',), "RFC 6087: 4.12"),
    'leaf-list':(('description',), "RFC 6087: 4.12"),
    'list':(('description',), "RFC 6087: 4.12"),
    'choice':(('description',), "RFC 6087: 4.12"),
    'anyxml':(('description',), "RFC 6087: 4.12"),
    }

_recommended_substatements = {
    'enum':(('description',), "RFC 6087: 4.10,4.12"),
    'bit':(('description',), "RFC 6087: 4.10,4.12"),
    }

_ietf_namespace_prefix = 'urn:ietf:params:xml:ns:yang:'

def v_chk_default(ctx, stmt):
    if (stmt.arg == _keyword_with_default[stmt.keyword] and
        stmt.parent.keyword != 'refine'):
        err_add(ctx.errors, stmt.pos, 'LINT_EXPLICIT_DEFAULT',
                (stmt.keyword, stmt.arg))

def v_chk_required_substmt(ctx, stmt):
    if stmt.keyword in _required_substatements:
        (required, s) = _required_substatements[stmt.keyword]
        for r in required:
            if stmt.search_one(r) is None:
                err_add(ctx.errors, stmt.pos,
                        'LINT_MISSING_REQUIRED_SUBSTMT',
                        (s, stmt.keyword, r))

def v_chk_recommended_substmt(ctx, stmt):
    if stmt.keyword in _recommended_substatements:
        (recommended, s) = _recommended_substatements[stmt.keyword]
        for r in recommended:
            if stmt.search_one(r) is None:
                err_add(ctx.errors, stmt.pos,
                        'LINT_MISSING_RECOMMENDED_SUBSTMT',
                        (s, stmt.keyword, r))

def v_chk_namespace(ctx, stmt, namespace_prefixes):
    if namespace_prefixes != []:
        for prefix in namespace_prefixes:
            if stmt.arg == prefix + stmt.i_module.arg:
                return
        err_add(ctx.errors, stmt.pos, 'LINT_BAD_NAMESPACE_VALUE',
                namespace_prefixes[0] + stmt.i_module.arg)

def v_chk_module_name(ctx, stmt, modulename_prefixes):
    if modulename_prefixes != []:
        for prefix in modulename_prefixes:
            if stmt.arg.find(prefix + '-') == 0:
                return
        if len(modulename_prefixes) == 1:
            err_add(ctx.errors, stmt.pos, 'LINT_BAD_MODULENAME_PREFIX_1',
                    '"' + modulename_prefixes[0] + '-"')
        elif len(modulename_prefixes) == 2:
            s = " or ".join(['"' + p + '-"' for p in modulename_prefixes])
            err_add(ctx.errors, stmt.pos, 'LINT_BAD_MODULENAME_PREFIX_N', s)
        else:
            s = ", ".join(['"' + p + '-"' for p in modulename_prefixes[:-1]]) +\
            ', or "' + modulename_prefixes[-1] + '-"'
            err_add(ctx.errors, stmt.pos, 'LINT_BAD_MODULENAME_PREFIX_N', s)
    elif stmt.arg.find('-') == -1:
        # can't check much, but we can check that a prefix is used
        err_add(ctx.errors, stmt.pos, 'LINT_NO_MODULENAME_PREFIX', ())

def v_chk_include(ctx, stmt):
    if stmt.i_orig_module.keyword != 'module':
        # the rule applies only to modules
        return
    latest = stmt.i_orig_module.i_latest_revision
    if latest is None:
        return
    submodulename = stmt.arg
    r = stmt.search_one('revision-date')
    if r is not None:
        rev = r.arg
    else:
        rev = None
    subm = ctx.get_module(submodulename, rev)
    if (subm is not None and
        subm.i_latest_revision is not None and
        subm.i_latest_revision > latest):
        err_add(ctx.errors, stmt.pos, 'LINT_BAD_REVISION',
                (latest, submodulename, subm.i_latest_revision))

def v_chk_mandatory_top_level(ctx, stmt):
    for s in stmt.i_children:
        if statements.is_mandatory_node(s):
            err_add(ctx.errors, s.pos, 'LINT_TOP_MANDATORY', s.arg)



"""YANG built-in types"""

from .error import err_add
from . import util
from . import syntax
import base64
from xml.sax.saxutils import quoteattr
from xml.sax.saxutils import escape

try:
    # python 2
    from StringIO import StringIO
except ImportError:
    # python 3
    from io import StringIO

class Abort(Exception):
    pass

class TypeSpec(object):
    def __init__(self, name):
        self.definition = ""
        self.name = name
        self.base = None

    def str_to_val(self, errors, pos, str):
        return str;

    def validate(self, errors, pos, val, errstr=''):
        return True;

    def restrictions(self):
        return []

class IntTypeSpec(TypeSpec):
    def __init__(self, name, min, max):
        TypeSpec.__init__(self, name)
        self.min = min
        self.max = max

    def str_to_val(self, errors, pos, str):
        try:
            if str in ['min', 'max']:
                return str
            return int(str, 0)
        except ValueError:
            err_add(errors, pos, 'TYPE_VALUE',
                    (str, self.definition, 'not an integer'))
            return None

    def validate(self, errors, pos, val, errstr = ''):
        if val < self.min or val > self.max:
            err_add(errors, pos, 'TYPE_VALUE',
                    (str(val), self.definition, 'range error' + errstr))
            return False
        else:
            return True

    def restrictions(self):
        return ['range']

class Decimal64Value(object):
    def __init__(self, value, s=None, fd=None):
        # must set s (string repr) OR fd (fraction-digits)
        self.value = value
        self.s = s
        if s == None and fd is not None:
            s = str(value)
            self.s = s[:-fd] + "." + s[-fd:]

    def __str__(self):
        return self.s

    def __cmp__(self, other):
        if not isinstance(other, Decimal64Value):
            return -1
        if self.value < other.value:
            return -1;
        elif self.value == other.value:
            return 0;
        else:
            return 1

    def __eq__(self, other):
        if not isinstance(other, Decimal64Value):
            return False
        return self.value == other.value

    def __ne__(self, other):
        if not isinstance(other, Decimal64Value):
            return True
        return self.value != other.value

    def __lt__(self, other):
        if not isinstance(other, Decimal64Value):
            return True
        return self.value < other.value

    def __le__(self, other):
        if not isinstance(other, Decimal64Value):
            return True
        return self.value <= other.value

    def __gt__(self, other):
        if not isinstance(other, Decimal64Value):
            return False
        return self.value > other.value

    def __ge__(self, other):
        if not isinstance(other, Decimal64Value):
            return False
        return self.value >= other.value


class Decimal64TypeSpec(TypeSpec):
    def __init__(self, fraction_digits):
        TypeSpec.__init__(self, 'decimal64')
        self.fraction_digits = int(fraction_digits.arg)
        self.min = Decimal64Value(-9223372036854775808, fd=self.fraction_digits)
        self.max = Decimal64Value(9223372036854775807, fd=self.fraction_digits)

    def str_to_val(self, errors, pos, s0):
        if s0 in ('min', 'max'):
            return s0
        # make sure it is syntactically correct
        if syntax.re_decimal.search(s0) is None:
            err_add(errors, pos, 'TYPE_VALUE',
                    (s0, self.definition, 'not a decimal'))
            return None
        if s0[0] == '-':
            is_negative = True
            s = s0[1:]
        else:
            is_negative = False
            s = s0
        p = s.find('.')
        if p == -1:
            v = int(s)
            i = self.fraction_digits
            while i > 0:
                v = v * 10
                i -= 1
        else:
            v = int(s[:p])
            i = self.fraction_digits
            j = p + 1
#            slen = len(s.rstrip('0')) # ignore trailing zeroes
# No, do not ignore trailing zeroes!
            slen = len(s)
            while i > 0:
                v *= 10
                i -= 1
                if j < slen:
                    v += int(s[j])
                j += 1
            if j < slen:
                err_add(errors, pos, 'TYPE_VALUE',
                        (s, self.definition, 'too many fraction digits'))
                return None
        if is_negative:
            v = -v
        return Decimal64Value(v, s=s0)

    def validate(self, errors, pos, val, errstr = ''):
        if val < self.min or val > self.max:
            err_add(errors, pos, 'TYPE_VALUE',
                    (str(val), self.definition, 'range error' + errstr))
            return False
        else:
            return True

    def restrictions(self):
        return ['range']

class BooleanTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'boolean')

    def str_to_val(self, errors, pos, str):
        if str == 'true': return True;
        elif str == 'false': return False
        else:
            err_add(errors, pos, 'TYPE_VALUE',
                    (str, self.definition, 'not a boolean'))
            return None

class StringTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'string')

    def restrictions(self):
        return ['pattern', 'length']

class BinaryTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'binary')

    def str_to_val(self, errors, pos, s):
        try:
            return base64.b64decode(s)
        except:
            err_add(errors, pos, 'TYPE_VALUE',
                    (s, '', 'bad base64 value'))

    def restrictions(self):
        return ['length']

class EmptyTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'empty')

    def str_to_val(self, errors, pos, str):
        err_add(errors, pos, 'BAD_DEFAULT_VALUE', 'empty')
        return None

class IdentityrefTypeSpec(TypeSpec):
    def __init__(self, idbases):
        TypeSpec.__init__(self, 'identityref')
        self.idbases = idbases

    def str_to_val(self, errors, pos, s):
        if s.find(":") == -1:
            prefix = None
            name = s
        else:
            [prefix, name] = s.split(':', 1)
        if prefix is None or self.idbases[0].i_module.i_prefix == prefix:
            # check local identities
            pmodule = self.idbases[0].i_module
        else:
            # this is a prefixed name, check the imported modules
            pmodule = util.prefix_to_module(self.idbases[0].i_module, prefix,
                                            pos, errors)
            if pmodule is None:
                return None
        if name not in pmodule.i_identities:
            err_add(errors, pos, 'TYPE_VALUE',
                    (s, self.definition, 'identityref not found'))
            return None
        val = pmodule.i_identities[name]
        for idbase in self.idbases:
            my_identity = idbase.i_identity
            if not is_derived_from(val, my_identity):
                err_add(errors, pos, 'TYPE_VALUE',
                        (s, self.definition,
                         'identityref not derived from %s' % \
                         my_identity.arg))
                return None
        else:
            return val

def is_derived_from(a, b):
    if a == b:
        # an identity is not derived from itself
        return False
    else:
        return is_derived_from_or_self(a, b, [])

def is_derived_from_or_self(a, b, visited):
    # return True if a is derived from b
    if a == b:
        return True
    for p in a.search('base'):
        val = p.i_identity
        if val not in visited:
            visited.append(val)
            if is_derived_from_or_self(val, b, visited):
                return True
    return False

## type restrictions

def validate_range_expr(errors, stmt, type_):
    # break the expression apart
    def f(lostr, histr):
        if histr == '':
            # this means that a single number was in the range, e.g.
            # "4 | 5..6".
            return (type_.i_type_spec.str_to_val(errors, stmt.pos, lostr),
                    None)
        return (type_.i_type_spec.str_to_val(errors, stmt.pos, lostr),
                type_.i_type_spec.str_to_val(errors, stmt.pos, histr))
    ranges = [f(m[1], m[6]) for m in syntax.re_range_part.findall(stmt.arg)]
    return validate_ranges(errors, stmt.pos, ranges, type_)

def validate_ranges(errors, pos, ranges, type_):
    # make sure the range values are of correct type and increasing
    cur_lo = None
    for (lo, hi) in ranges:
        if lo != 'min' and lo != 'max':
            type_.i_type_spec.validate(errors, pos, lo)
        if hi != 'min' and hi != 'max' and hi != None:
            type_.i_type_spec.validate(errors, pos, hi)
        # check that cur_lo < lo < hi
        if not is_smaller(cur_lo, lo):
            err_add(errors, pos, 'RANGE_BOUNDS', (str(lo), cur_lo))
            return None
        if not is_smaller(lo, hi):
            err_add(errors, pos, 'RANGE_BOUNDS', (str(hi), str(lo)))
            return None
        if hi == None:
            cur_lo = lo
        else:
            cur_lo = hi
    return (ranges, pos)

class RangeTypeSpec(TypeSpec):
    def __init__(self, base, range_spec):
        TypeSpec.__init__(self, base.name)
        self.base = base
        (ranges, ranges_pos) = range_spec
        self.ranges = ranges
        self.ranges_pos = ranges_pos
        if ranges != []:
            self.min = ranges[0][0]
            if self.min == 'min':
                self.min = base.min
            self.max = ranges[-1][1]
            if self.max == None: # single range
                self.max = ranges[-1][0]
            if self.max == 'max':
                self.max = base.max
        else:
            self.min = base.min
            self.max = base.max
        if hasattr(base, 'fraction_digits'):
            self.fraction_digits = base.fraction_digits

    def str_to_val(self, errors, pos, str):
        return self.base.str_to_val(errors, pos, str)

    def validate(self, errors, pos, val, errstr=''):
        if self.base.validate(errors, pos, val, errstr) == False:
            return False
        for (lo, hi) in self.ranges:
            if ((lo == 'min' or lo == 'max' or val >= lo) and
                ((hi is None and val == lo) or hi == 'max' or \
                     (hi is not None and val <= hi))):
                return True
        err_add(errors, pos, 'TYPE_VALUE',
                (str(val), self.definition, 'range error' + errstr +
                 ' for range defined at ' + str(self.ranges_pos)))
        return False

    def restrictions(self):
        return self.base.restrictions()

def validate_length_expr(errors, stmt):
    def f(lostr, histr):
        try:
            if lostr in ['min', 'max']:
                lo = lostr
            else:
                lo = int(lostr)
        except ValueError:
            err_add(errors, stmt.pos, 'TYPE_VALUE',
                    (lostr, '', 'not an integer'))
            return (None, None)
        try:
            if histr == '':
                # this means that a single number was in the length, e.g.
                # "4 | 5..6".
                return (lo, None)
            if histr in ['min', 'max']:
                hi = histr
            else:
                hi = int(histr)
        except ValueError:
            err_add(errors, stmt.pos, 'TYPE_VALUE',
                    (histr, '', 'not an integer'))
            return None
        return (lo, hi)
    lengths = [f(m[1], m[3]) for m in syntax.re_length_part.findall(stmt.arg)]
    # make sure the length values are of correct type and increasing
    cur_lo = None
    for (lo, hi) in lengths:
        # check that cur_lo < lo < hi
        if not is_smaller(cur_lo, lo):
            err_add(errors, stmt.pos, 'LENGTH_BOUNDS', (str(lo), cur_lo))
            return None
        if not is_smaller(lo, hi):
            err_add(errors, stmt.pos, 'LENGTH_BOUNDS', (str(hi), str(lo)))
            return None
        # FIXME: we should check that the lengths are just restrictions
        # of any base type's lengths.  Have to figure out some way to do
        # that... currently we can't check just length values; we'd have
        # to pass just the length integer to typespec.validate().  Or
        # something...
        if hi == None:
            cur_lo = lo
        else:
            cur_lo = hi
        if type(cur_lo) == type(0) and cur_lo > 18446744073709551615:
            err_add(errors, stmt.pos, 'LENGTH_VALUE', str(cur_lo))
            return None
    return (lengths, stmt.pos)

class LengthTypeSpec(TypeSpec):
    def __init__(self, base, length_spec):
        TypeSpec.__init__(self, base.name)
        self.base = base
        (lengths, length_pos) = length_spec
        self.lengths = lengths
        self.length_pos = length_pos

    def str_to_val(self, errors, pos, str):
        return self.base.str_to_val(errors, pos, str)

    def validate(self, errors, pos, val, errstr=''):
        if self.base.validate(errors, pos, val, errstr) == False:
            return False
        vallen = len(val)
        for (lo, hi) in self.lengths:
            if ((lo == 'min' or vallen >= lo) and
                ((hi is None and vallen == lo) or hi == 'max' or
                 (hi is not None and vallen <= hi))):
                return True
        err_add(errors, pos, 'TYPE_VALUE',
                (val, self.definition, 'length error' + errstr +
                 ' for length defined at ' + str(self.length_pos)))
        return False

    def restrictions(self):
        return self.base.restrictions()


def _validate_pattern_libxml2(errors, stmt, invert_match):
    try:
        import libxml2
        try:
            re = libxml2.regexpCompile(stmt.arg)
            return ('libxml2', re, stmt.pos, invert_match)
        except libxml2.treeError as v:
            err_add(errors, stmt.pos, 'PATTERN_ERROR', str(v))
            return None
    except ImportError:
    ## Do not report a warning in this case.  Maybe we should add some
    ## flag to turn on this warning...
    #        err_add(errors, stmt.pos, 'PATTERN_FAILURE',
    #                "Could not import python module libxml2 "
    #                    "(see http://xmlsoft.org for installation help)")
        return False

def _validate_pattern_lxml(errors, stmt, invert_match):
    try:
        import lxml.etree
        doc = StringIO(
            '<xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema">' \
            '  <xsd:element name="a" type="x"/>' \
            '    <xsd:simpleType name="x">' \
            '      <xsd:restriction base="xsd:string">' \
            '        <xsd:pattern value=%s/>' \
            '      </xsd:restriction>' \
            '     </xsd:simpleType>' \
            '   </xsd:schema>' % quoteattr(stmt.arg))
        try:
            sch = lxml.etree.XMLSchema(lxml.etree.parse(doc))
            return ('lxml', sch, stmt.pos, invert_match)
        except lxml.etree.XMLSchemaParseError as v:
            err_add(errors, stmt.pos, 'PATTERN_ERROR', str(v))
            return None
    except ImportError:
        return False

def validate_pattern_expr(errors, stmt):
    invert_match = False
    if stmt.search_one('modifier', arg='invert-match') is not None:
        invert_match = True
    ## check that it's syntactically correct
    # First try with lxml
    res = _validate_pattern_lxml(errors, stmt, invert_match)
    if res is not False:
        return res
    # Then try with libxml2
    res = _validate_pattern_libxml2(errors, stmt, invert_match)
    if res is not False:
        return res
    # Otherwise we can't validate patterns :(
    return None

class PatternTypeSpec(TypeSpec):
    def __init__(self, base, pattern_specs):
        TypeSpec.__init__(self, base.name)
        self.base = base
        self.res = pattern_specs

    def str_to_val(self, errors, pos, str):
        return self.base.str_to_val(errors, pos, str)

    def validate(self, errors, pos, val, errstr=''):
        if self.base.validate(errors, pos, val, errstr) == False:
            return False
        for (type_, re, re_pos, invert_match) in self.res:
            if type_ == 'libxml2':
                is_valid = re.regexpExec(val) == 1
            elif type_ == 'lxml':
                import lxml
                doc = StringIO('<a>%s</a>' % escape(val))
                is_valid = re.validate(lxml.etree.parse(doc))
            if ((not is_valid and not invert_match) or
                (is_valid and invert_match)):
                err_add(errors, pos, 'TYPE_VALUE',
                        (val, self.definition, 'pattern mismatch' + errstr +
                         ' for pattern defined at ' + str(re_pos)))
                return False
        return True

    def restrictions(self):
        return self.base.restrictions()

def validate_enums(errors, enums, stmt):
    # make sure all names and values given are unique
    names = {}
    values = {}
    next = 0
    for e in enums:
        # for derived enumerations, make sure the enum is defined
        # in the base
        stmt.i_type_spec.validate(errors, e.pos, e.arg)
        e.i_value = None
        value = e.search_one('value')
        if value is not None:
            try:
                x = int(value.arg)
                # for derived enumerations, make sure the value isn't changed
                oldval = stmt.i_type_spec.get_value(e.arg)
                if oldval is not None and oldval != x:
                    err_add(errors, value.pos, 'BAD_ENUM_VALUE',
                            (value.arg, oldval))
                e.i_value = x
                if x < -2147483648 or x > 2147483647:
                    raise ValueError
                if x >= next:
                    next = x + 1
                if x in values:
                    err_add(errors, value.pos, 'DUPLICATE_ENUM_VALUE',
                            (x, values[x]))
                else:
                    values[x] = value.pos

            except ValueError:
                err_add(errors, value.pos, 'ENUM_VALUE', value.arg)
        else:
            # auto-assign a value
            values[next] = e.pos
            if next > 2147483647:
                err_add(errors, e.pos, 'ENUM_VALUE', str(next))
            e.i_value = next
            next = next + 1
        if e.arg in names:
            err_add(errors, e.pos, 'DUPLICATE_ENUM_NAME', (e.arg, names[e.arg]))
        else:
            names[e.arg] = e.pos

    # check status (here??)
    return enums

class EnumerationTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'enumeration')

    def get_value(self, val):
        return None

    def restrictions(self):
        return ['enum']

class EnumTypeSpec(TypeSpec):
    def __init__(self, base, enums):
        TypeSpec.__init__(self, base.name)
        self.base = base
        self.enums = [(e.arg, e.i_value) for e in enums]

    def validate(self, errors, pos, val, errstr = ''):
        if util.keysearch(val, 0, self.enums) == None:
            err_add(errors, pos, 'TYPE_VALUE',
                    (val, self.definition, 'enum not defined' + errstr))
            return False
        else:
            return True

    def get_value(self, val):
        r  = util.keysearch(val, 0, self.enums)
        if r is not None:
            return r[1]
        else:
            return None

    def restrictions(self):
        return self.base.restrictions()

def validate_bits(errors, bits, stmt):
    # make sure all names and positions given are unique
    names = {}
    values = {}
    next = 0
    for b in bits:
        # for derived bits, make sure the bit is defined
        # in the base
        stmt.i_type_spec.validate(errors, b.pos, [b.arg])
        position = b.search_one('position')
        if position is not None:
            try:
                x = int(position.arg)
                # for derived bits, make sure the position isn't changed
                oldpos = stmt.i_type_spec.get_position(b.arg)
                if oldpos is not None and oldpos != x:
                    err_add(errors, position.pos, 'BAD_BIT_POSITION',
                            (position.arg, oldpos))
                b.i_position = x
                if x < 0 or x > 4294967295:
                    raise ValueError
                if x >= next:
                    next = x + 1
                if x in values:
                    err_add(errors, position.pos, 'DUPLICATE_BIT_POSITION',
                            (x, values[x]))
                else:
                    values[x] = position.pos
            except ValueError:
                err_add(errors, position.pos, 'BIT_POSITION', position.arg)
        else:
            # auto-assign a value
            values[next] = b.pos
            b.i_position = next
            next = next + 1
        if b.arg in names:
            err_add(errors, b.pos, 'DUPLICATE_BIT_NAME', (b.arg, names[b.arg]))
        else:
            names[b.arg] = b.pos

    # check status (here??)
    return bits

class BitsTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'bits')

    def get_position(self, bit):
        return None

    def restrictions(self):
        return ['bit']

class BitTypeSpec(TypeSpec):
    def __init__(self, base, bits):
        TypeSpec.__init__(self, base.name)
        self.base = base
        self.bits = [(b.arg, b.i_position) for b in bits]

    def str_to_val(self, errors, pos, str):
        return str.split()

    def validate(self, errors, pos, val, errstr = ''):
        for v in val:
            if util.keysearch(v, 0, self.bits) == None:
                err_add(errors, pos, 'TYPE_VALUE',
                        (v, self.definition, 'bit not defined' + errstr))
                return False
        return True

    def get_position(self, bit):
        r  = util.keysearch(bit, 0, self.bits)
        if r is not None:
            return r[1]
        else:
            return None

    def restrictions(self):
        return self.base.restrictions()

def validate_path_expr(errors, path):

    # FIXME: rewrite using the new xpath tokenizer

    # PRE: s matches syntax.path_arg
    # -type dn [identifier | ('predicate', identifier, up::int(), [identifier])]
    # Ret: (up::int(),
    #       dn::dn(),
    #       derefup::int(),
    #       derefdn::dn())
    def parse_keypath(s):

        def parse_dot_dot(s):
            up = 0
            i = 0
            while True:
                if s[i] == '.' and s[i+1] == '.':
                    up = up + 1
                    i = i + 3 # skip the '/'
                elif s[i] == '/':
                    i = i + 1 # skip the '/'
                    if up == 0: # absolute path
                        up = -1
                    break
                elif s[i].isspace():
                    i = i + 1
                else:
                    # s points to an identifier
                    break
            return (up, s[i:])

        def skip_space(s):
            if len(s) == 0:
                return s
            i = 0
            while s[i].isspace():
                i = i + 1
            return s[i:]

        def parse_identifier(s):
            m = syntax.re_keyword_start.match(s)
            if m is None:
                raise Abort
            s = s[m.end():]
            if m.group(2) is None:
                # no prefix
                return (m.group(3), s)
            else:
                prefix = m.group(2)
                mod = util.prefix_to_module(path.i_module, prefix,
                                            path.pos, errors)
                if mod is not None:
                    return ((m.group(2), m.group(3)), s)
                else:
                    raise Abort

        def parse_key_predicate(s):
            s = s[1:] # skip '['
            s = skip_space(s)
            (identifier, s) = parse_identifier(s)
            s = skip_space(s)
            s = s[1:] # skip '='
            s = skip_space(s)
            if s[:7] == 'current':
                s = s[7:] # skip 'current'
                s = skip_space(s)
                s = s[1:] # skip '('
                s = skip_space(s)
                s = s[1:] # skip ')'
                s = skip_space(s)
                s = s[1:] # skip '/'
                s = skip_space(s)
                (up, s) = parse_dot_dot(s)
                s = skip_space(s)
            else:
                up = -1
                b = s.find(']') + 1
                s = s[b:]
                if len(s) > 0 and s[0] == '/':
                    s = s[1:] # skip '/'
            dn = []
            while len(s) > 0:
                (xidentifier, s) = parse_identifier(s)
                dn.append(xidentifier)
                s = skip_space(s)
                if len(s) == 0:
                    break
                if s[0] == '/':
                    s = s[1:] # skip '/'
                elif s[0] == ']':
                    s = s[1:] # skip ']'
                    break
            return (('predicate', identifier, up, dn), s)

        def parse_descendant(s):
            dn = []
            # all '..'s are now parsed
            while len(s) > 0 and (not s[0].isspace()) and s[0] != ')':
                (identifier, s) = parse_identifier(s)
                dn.append(identifier)
                s = skip_space(s)
                if len(s) == 0:
                    break
                while len(s) > 0 and s[0] == '[':
                    (pred, s) = parse_key_predicate(s)
                    dn.append(pred)
                    s = skip_space(s)
                if len(s) > 0 and s[0] == '/':
                    s = s[1:] # skip '/'

            return (dn, s)

        derefup = 0
        derefdn = None
        if s.startswith('deref'):
            s = s[5:] # skip 'deref'
            s = skip_space(s)
            s = s[1:] # skip '('
            s = skip_space(s)
            (derefup, s) = parse_dot_dot(s)
            (derefdn, s) = parse_descendant(s)
            s = skip_space(s)
            s = s[1:] # skip ')'
            s = skip_space(s)
            s = s[1:] # skip '/'

        (up, s) = parse_dot_dot(s)
        (dn, s) = parse_descendant(s)
        return (up, dn, derefup, derefdn)

    try:
        return parse_keypath(path.arg)
    except Abort:
        return None

class LeafrefTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'leafref')
        self.require_instance = True

    def restrictions(self):
        return ['path', 'require-instance']

class InstanceIdentifierTypeSpec(TypeSpec):
    def __init__(self):
        TypeSpec.__init__(self, 'instance-identifier')
        self.require_instance = True

    def restrictions(self):
        return ['require-instance']

class PathTypeSpec(TypeSpec):
    def __init__(self, base, path_spec, path, pos):
        TypeSpec.__init__(self, base.name)
        self.require_instance = True
        self.base = base
        self.path_spec = path_spec
        self.path_ = path
        self.pos = pos

    def str_to_val(self, errors, pos, str_):
        if hasattr(self, 'i_target_node'):
            return self.i_target_node.search_one('type').\
                i_type_spec.str_to_val(errors, pos, str_)
        else:
            # if a default value is verified
            return str_

    def validate(self, errors, pos, val, errstr = ''):
        if hasattr(self, 'i_target_node'):
            return self.i_target_node.search_one('type').\
                i_type_spec.validate(errors, pos, val)
        else:
            # if a default value is verified
            return True

    def restrictions(self):
        return ['require-instance']

class UnionTypeSpec(TypeSpec):
    def __init__(self, types):
        TypeSpec.__init__(self, 'union')
        # no base - no restrictions allowed
        self.types = types

    def str_to_val(self, errors, pos, str):
        return str

    def validate(self, errors, pos, str, errstr = ''):
        # try to validate against each membertype
        for t in self.types:
            if t.i_type_spec != None:
                val = t.i_type_spec.str_to_val([], pos, str)
                if val != None:
                    if t.i_type_spec.validate([], pos, val):
                        return True;
        err_add(errors, pos, 'TYPE_VALUE',
                (str, self.definition, 'no member type matched' + errstr))
        return False

yang_type_specs = \
  {'int8':IntTypeSpec('int8', -128, 127),
   'int16':IntTypeSpec('int16', -32768, 32767),
   'int32':IntTypeSpec('int32', -2147483648, 2147483647),
   'int64':IntTypeSpec('int64', -9223372036854775808, 9223372036854775807),
   'uint8':IntTypeSpec('uint8', 0, 255),
   'uint16':IntTypeSpec('uint16', 0, 65535),
   'uint32':IntTypeSpec('uint32', 0, 4294967295),
   'uint64':IntTypeSpec('uint64', 0, 18446744073709551615),
   'decimal64':TypeSpec('decimal64'),
   'string':StringTypeSpec(),
   'boolean':BooleanTypeSpec(),
   'enumeration':EnumerationTypeSpec(),
   'bits':BitsTypeSpec(),
   'binary':BinaryTypeSpec(),
   'leafref':LeafrefTypeSpec(),
   'identityref':TypeSpec('identityref'),
   'instance-identifier':InstanceIdentifierTypeSpec(),
   'empty':EmptyTypeSpec(),
   'union':TypeSpec('union'),
   }

def is_base_type(typename):
    return typename in yang_type_specs

def is_smaller(lo, hi):
    if lo == None:
        return True;
    if lo == 'min' and hi != 'min':
        return True;
    if lo == 'max' and hi != None:
        return False
    if hi == 'min':
        return False
    if hi == None:
        return True
    if hi == 'max':
        return True
    return lo < hi

